#include <sqlextract/strid.h>
#include <sqlextract/util.h>
#include <string.h>
#include <stdlib.h>

struct strid_tree_t {
	int id;
	char *str;
	struct strid_tree_t *lo;
	struct strid_tree_t *hi;
};

struct strid_t {
	int next_id;
	struct strid_tree_t *root;
};

static void strid_tree_destroy(struct strid_tree_t *sit);

struct strid_t *strid_create(void) {
	struct strid_t *si;

	si = (struct strid_t *) lf_calloc(1, sizeof(struct strid_t));
	si->next_id = 0;

	return si;
}

static void strid_tree_destroy(struct strid_tree_t *sit) {
	if (sit == NULL)
		return;
	strid_tree_destroy(sit->lo);
	strid_tree_destroy(sit->hi);
	free(sit->str);
	free(sit);
}

void strid_destroy(struct strid_t *si) {
	if (si == NULL)
		return;
	strid_tree_destroy(si->root);
	free(si);
}

int strid_get(struct strid_t *si, const char *str) {
	int id;
	int diff;
	struct strid_tree_t **cur;

	/*
	 * Try to find the node, but keep track of the pointer to where to store the
	 * node, if not found, so we can add it later
	 */
	cur = &(si->root);
	while (*cur != NULL) {
		diff = strcmp(str, (*cur)->str);
		if (diff < 0) {
			cur = &((*cur)->lo);
		} else if (diff > 0) {
			cur = &((*cur)->hi);
		} else {
			/* Found it, just return the id */
			return (*cur)->id;
		}
	}

	/*
	 * No strings are found. Generate and save a new node, if possible.
	 */

	/* Generate a new id */
	id = si->next_id;
	si->next_id++;

	/* And save the node */
	*cur = lf_calloc(1, sizeof(struct strid_tree_t));
	(*cur)->str = lf_strdup(str);
	(*cur)->id = id;
	(*cur)->lo = NULL;
	(*cur)->hi = NULL;
	return (*cur)->id;
}
