#ifndef TABLE_H_
#define TABLE_H_

#include <sqlextract/filter/arbval.h>
#include <sqlextract/filter/filter.h>
#include <sqlextract/strid.h>

struct table_t;
struct table_row_t;

struct table_desc_t {
	/* Destroys a table storage */
	void *(*create_table)(struct strid_t *, void *);

	/* Destroys a table storage */
	void (*destroy_table)(void *);

	/* Create aa iterator object, takes a table storage as argument (passed to table_create) */
	void *(*create_iterator)(void *);

	/* Fetch a row from the table, the iterator storage. Return non-zero on valid row, 0 otherwise */
	int (*fetch_row)(void *);

	/* Fetch a field the current row, given an iterator */
	struct arbval_t *(*fetch_field)(void *, int);

	/* Destroys an iterator object (argument 1), but also get a table storage as second parameter */
	void (*destroy_iterator)(void *);
};

struct table_t *table_create(const struct table_desc_t *tdesc, struct strid_t *si, void *arg);
void table_destroy(struct table_t *tbl);

void table_filter(struct table_t *tbl, int (*cb)(struct table_row_t *, void *),
		void *arg, struct filter_t *flt);

struct arbval_t *table_row_fetch_column(struct table_row_t *row, int fieldname);

struct filter_t *filter_create_table_column(int fieldname);

#endif /* TABLE_H_ */
